@@ -1,8 +1,10 @@
-use Feature::Compat::Class 0.04;
-
-use v5.12;
+use v5.20;
 use utf8;
 use warnings;
+use feature qw(signatures);
+no warnings qw(experimental::signatures);
+
+use Feature::Compat::Class 0.04;
 
 =head1 NAME
 
@@ -10,7 +12,7 @@ String::License - detect source code license statements in a text string
 
 =head1 VERSION
 
-Version v0.0.5
+Version v0.0.6
 
 =head1 SYNOPSIS
 
@@ -33,7 +35,7 @@ and serializes them in a normalized format.
 
 =cut
 
-package String::License v0.0.5;
+package String::License v0.0.6;
 
 use Carp            qw(croak);
 use Log::Any        ();
@@ -47,137 +49,27 @@ use String::License::Naming::SPDX;
 
 use namespace::clean;
 
-class Trait {
-	field $log;
-	field $name :param;
-	field $begin :param;
-	field $end :param;
-	field $file :param;
-
-	ADJUST {
-		$log = Log::Any->get_logger;
-
-		$log->tracef(
-			'located trait: %s: %d-%d "%s"',
-			$name, $begin, $end,
-			$file ? substr( $file->string, $begin, $end - $begin ) : '',
-		);
-	}
-
-	method name { return $name }
-	method begin { return $begin }
-	method end { return $end }
-	method file { return $file }
-}
-
-class Exception {
-	field $log;
+class Tag {
 	field $id :param;
 	field $begin :param;
 	field $end :param;
-	field $file :param;
 
-	ADJUST {
-		$log = Log::Any->get_logger;
-
-		$log->tracef(
-			'detected exception: %s: %d-%d',
-			$id->{caption}, $begin, $end
-		);
-	}
-
-	method id { return $id }
-	method begin { return $begin }
-	method end { return $end }
-	method file { return $file }
+	method data () { return lc __CLASS__, $id, $begin, $end }
 }
 
-class Flaw {
-	field $log;
-	field $id :param;
-	field $begin :param;
-	field $end :param;
-	field $file :param;
+class Exception :isa(Tag) { }
 
-	ADJUST {
-		$log = Log::Any->get_logger;
-
-		$log->tracef(
-			'detected flaw: %s: %d-%d',
-			$id->{caption}, $begin, $end
-		);
-	}
-
-	method id { return $id }
-	method begin { return $begin }
-	method end { return $end }
-	method file { return $file }
-}
+class Flaw :isa(Tag) { }
 
 class Licensing {
-	field $log;
-	field $name :param;
+	field $id :param;
 
-	ADJUST {
-		$log = Log::Any->get_logger;
-
-		$log->debugf(
-			'collected some licensing: %s',
-			$name
-		);
-	}
-
-	method name { return $name }
+	method data () { return lc __CLASS__, $id }
 }
 
-class Fulltext {
-	field $log;
-	field $name :param;
-	field $begin :param;
-	field $end :param;
-	field $file :param;
-	field $traits :param = undef;
+class Fulltext :isa(Tag) { }
 
-	ADJUST {
-		$log = Log::Any->get_logger;
-
-		$log->debugf(
-			'collected fulltext: %s: %d-%d',
-			$name, $begin, $end
-		);
-	}
-
-	method name { return $name }
-	method begin { return $begin }
-	method end { return $end }
-	method file { return $file }
-	method traits { return $traits }
-}
-
-class Grant {
-	field $log;
-	field $name :param;
-	field $begin :param;
-	field $end :param;
-	field $file :param;
-	field $traits :param = undef;
-
-	ADJUST {
-		$log = Log::Any->get_logger;
-
-		$log->debugf(
-			'collected grant: %s: %d-%d "%s"',
-			$name, $begin, $end,
-			$file ? substr( $file->string, $begin, $end - $begin ) : '',
-		);
-	}
-
-	method name { return $name }
-	method begin { return $begin }
-	method end { return $end }
-	method file { return $file }
-	method traits { return $file }
-}
+class Grant :isa(Tag) { }
 
 class String::License;
 
@@ -229,6 +121,9 @@ field $naming :param = undef;
 field $license = '';
 field $expr = '';
 
+field $coverage;
+field @loose_licensing;
+
 =back
 
 =back
@@ -247,6 +142,8 @@ ADJUST {
 	else {
 		$naming = String::License::Naming::SPDX->new;
 	}
+
+	$coverage = Array::IntSpan->new();
 }
 
 =head1 METHODS
@@ -255,7 +152,82 @@ ADJUST {
 
 =cut
 
-method string
+method note ( $name, $begin, $end )
+{
+	$log->tracef(
+		'noted %s: %d-%d "%s"',
+		$name, $begin, $end,
+		substr( $string, $begin, $end - $begin )
+	);
+
+	return [ $begin, $end ];
+}
+
+method tag ($obj)
+{
+	my ( $name, $desc, $type, $id, $begin, $end );
+
+	( $type, $id, $begin, $end ) = $obj->data;
+
+	if ( ref($id) ) {
+		( $name, $desc ) = @$id;
+	}
+	else {
+		$name
+			= $self->best_value( $Regexp::Pattern::License::RE{$id}, 'name' );
+		$desc = $self->best_value(
+			$Regexp::Pattern::License::RE{$id},
+			qw(caption name)
+		);
+	}
+
+	if ( $type eq 'licensing' ) {
+		push @loose_licensing, [ $type, $name, $desc ];
+	}
+	else {
+		$coverage->set_range( $begin, $end, [ $type, $name, $desc ] );
+		$log->tracef(
+			'tagged %s: %s: %d-%d',
+			$type, $desc, $begin, $end
+		);
+	}
+
+	return $obj;
+}
+
+method contains_tag ( $begin, $end )
+{
+	return defined( $coverage->get_range( $begin, $end )->get_element(0) );
+}
+
+method get_tags ()
+{
+	my ( @thing, %set );
+
+	@thing = $coverage->get_range_list;
+	$set{grant}{ $_->[1] } = $_ for @loose_licensing;
+	for my $i ( 0 .. $#thing ) {
+		my ( $begin, $end, $thing, $type, $key );
+
+		( $begin, $end, $thing ) = $coverage->get_element($i);
+		$type = $thing->[0];
+
+		# TODO: drop fallback when all flaws have shortname
+		$key = $thing->[1] || $thing->[2];
+
+		next unless $type =~ /^[a-z]/;
+
+		$set{$type}{$key} = $thing;
+	}
+
+	return (
+		[ values %{ $set{fulltexts} }, values %{ $set{grant} } ],
+		[ values %{ $set{exception} } ],
+		[ values %{ $set{flaw} } ],
+	);
+}
+
+method string ()
 {
 	return $string;
 }
@@ -272,9 +244,8 @@ my $re_prop_attrs = qr/
 		(?'other'\.$any)
 	)*\z/x;
 
-method best_value
+method best_value ( $hashref, @props )
 {
-	my ( $hashref, @props ) = @_;
 	my $value;
 
 	PROPERTY:
@@ -304,7 +275,7 @@ my $type_re
 our %RE;
 my ( %L, @RE_EXCEPTION, @RE_LICENSE, @RE_NAME );
 
-method init_licensepatterns
+method init_licensepatterns ()
 {
 	# reuse if already resolved
 	return %L if exists $L{re_trait};
@@ -411,9 +382,9 @@ method init_licensepatterns
 		),
 	);
 
-	@RE_EXCEPTION = sort map /^EXCEPTION_(.*)/, keys(%RE);
-	@RE_LICENSE   = sort map /^LICENSE_(.*)/,   keys(%RE);
-	@RE_NAME      = sort map /^NAME_(.*)/,      keys(%RE);
+	@RE_EXCEPTION = sort map {/^EXCEPTION_(.*)/} keys(%RE);
+	@RE_LICENSE   = sort map {/^LICENSE_(.*)/} keys(%RE);
+	@RE_NAME      = sort map {/^NAME_(.*)/} keys(%RE);
 
 	foreach my $key ( grep {/^[a-z]/} keys(%Regexp::Pattern::License::RE) ) {
 		my $val = $Regexp::Pattern::License::RE{$key};
@@ -452,7 +423,6 @@ method init_licensepatterns
 	$L{multi_2} = qr/$RE{LOCAL_TRAIT_licensed_under}$RE{LOCAL_TRAIT_any_of}(?:[^.]|\.\S)*$RE{LOCAL_NAME_gpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
 	$L{lgpl_5} = qr/$RE{LOCAL_TRAIT_licensed_under}$RE{LOCAL_NAME_lgpl}(?:$RE{LOCAL_TRAIT_by_fsf})?[,;:]?(?: either)? ?$RE{LOCAL_TRAIT_KEEP_version_numberstring},? $RE{LOCAL_TRAIT_or_at_option} $RE{LOCAL_TRAIT_KEEP_version_numberstring}/i;
 	$L{gpl_7} = qr/either $RE{LOCAL_NAME_gpl}$RE{LOCAL_TRAIT_KEEP_version}?(?: \((?:the )?"?GPL"?\))?, or $RE{LOCAL_NAME_lgpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
-	$L{bsd_1} = qr/THIS SOFTWARE IS PROVIDED (?:BY (?:\S+ ){1,15})?AS IS AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY/;
 	$L{apache_1} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]+\))*,? or $RE{LOCAL_NAME_gpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
 	$L{apache_2} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or(?: the)? bsd(?:[ -](\d)-clause)?\b/i;
 	$L{apache_4} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or $RE{LOCAL_NAME_mit}\b/i;
@@ -508,16 +478,9 @@ my %L_grant_atomic_incomplete = (
 my $skip_stepwise = 0;
 my $force_atomic  = 0;
 
-my $contains_bsd2_re = qr/^license:contains:license:bsd_2_clause/;
-my @L_contains_bsd   = grep {
-	$Regexp::Pattern::License::RE{$_}{tags}
-		and grep /$contains_bsd2_re/,
-		@{ $Regexp::Pattern::License::RE{$_}{tags} }
-} keys(%Regexp::Pattern::License::RE);
-
 my $id2patterns_re = qr/(.*)(?:_(\d+(?:\.\d+)*)(_or_later)?)?/;
 
-method resolve
+method resolve ()
 {
 	$self->init_licensepatterns;
 
@@ -534,31 +497,28 @@ method resolve
 	my @gpl  = qw(gpl gpl_1 gpl_2 gpl_3);
 	my @lgpl = qw(lgpl lgpl_2 lgpl_2_1 lgpl_3);
 
-	my $coverage = Array::IntSpan->new();
 	my %match;
 	my ( %grant, %license );
 
-   # @clues, @expressions, and @exceptions contains DEP-5 or SPDX identifiers,
-   # and @flaws contains non-SPDX notes.
-	my ( @clues, @expressions, @exceptions, @flaws );
+	my $patterns2id = sub ( $stem, $version = undef ) {
+		return $stem
+			unless ($version);
+		$version =~ s/\.0$//g;
+		$version =~ s/\./_/g;
+		return "${stem}_$version";
+	};
+	my $id2patterns = sub ($id) {
+		return $id =~ /$id2patterns_re/;
+	};
+	my $gen_license = sub (
+		$id, $v = undef, $later = undef, $id2 = undef, $v2 = undef,
+		$later2 = undef
+		)
+	{
+		my ( @spdx, $name, $desc, $name2, $desc2, $legacy, $expr );
 
-	my $patterns2id = sub {
-		my ( $id, $ver ) = @_;
-		return $id
-			unless ($ver);
-		$_ = $ver;
-		s/\.0$//g;
-		s/\./_/g;
-		return "${id}_$_";
-	};
-	my $id2patterns = sub {
-		return $_[0] =~ /$id2patterns_re/;
-	};
-	my $gen_license = sub {
-		my ( $id, $v, $later, $id2, $v2, $later2 ) = @_;
-		my @spdx;
-		my $name = $L{name}{$id}    || $id;
-		my $desc = $L{caption}{$id} || $id;
+		$name = $L{name}{$id}    || $id;
+		$desc = $L{caption}{$id} || $id;
 		if ($v) {
 			push @spdx, $later ? "$name-$v+" : "$name-$v";
 			$v .= ' or later' if ($later);
@@ -566,7 +526,6 @@ method resolve
 		else {
 			push @spdx, $name;
 		}
-		my ( $name2, $desc2 );
 		if ($id2) {
 			$name2 = $L{name}{$id2}    || $id2;
 			$desc2 = $L{caption}{$id2} || $id2;
@@ -578,16 +537,19 @@ method resolve
 				push @spdx, $name2;
 			}
 		}
-		my $legacy = join(
+		$legacy = join(
 			' ',
 			$desc,
 			$v     ? "(v$v)"     : (),
 			$desc2 ? "or $desc2" : (),
 			$v2    ? "(v$v2)"    : (),
 		);
-		my $expr = join( ' or ', sort @spdx );
-		push @expressions, Licensing->new( name => $expr );
-		$license = join( ' ', $L{caption}{$legacy} || $legacy, $license );
+		$expr = join( ' or ', sort @spdx );
+		$self->tag(
+			Licensing->new(
+				id => [ $expr, $L{caption}{$legacy} || $legacy ]
+			)
+		);
 	};
 
 	# fulltext
@@ -596,12 +558,8 @@ method resolve
 	foreach my $id (@RE_LICENSE) {
 		next unless ( $RE{"LICENSE_$id"} );
 		while ( $string =~ /$RE{"LICENSE_$id"}/g ) {
-			$pos_license{ $-[0] }{$id} = Trait->new(
-				name  => "license($id)",
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
+			$pos_license{ $-[0] }{$id}
+				= $self->note( "license($id)", $-[0], $+[0] );
 		}
 	}
 
@@ -609,26 +567,15 @@ method resolve
 
 		next unless ( $string =~ /$RE{"TRAIT_$trait"}/ );
 		while ( $string =~ /$RE{"TRAIT_$trait"}/g ) {
-			next
-				if (
-				defined(
-					$coverage->get_range( $-[0], $+[0] )->get_element(0)
-				)
-				);
-			push @clues,
-				Trait->new(
-				name  => $trait,
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-				);
+			next if $self->contains_tag( $-[0], $+[0] );
+			$self->note( $trait, $-[0], $+[0] );
 		}
 	}
 	foreach my $pos ( sort { $a <=> $b } keys %pos_license ) {
 
 		# pick longest or most specific among matched license fulltexts
-		my @licenses = nsort_by { $pos_license{$pos}{$_}->end }
-		grep { $pos_license{$pos}{$_} ? $pos_license{$pos}{$_}->end : () } (
+		my @licenses = nsort_by { $pos_license{$pos}{$_}->[1] }
+		grep { $pos_license{$pos}{$_} ? $pos_license{$pos}{$_}->[1] : () } (
 			@L_type_group,
 			@L_type_combo,
 			@L_type_unversioned,
@@ -639,44 +586,30 @@ method resolve
 		my $license = pop @licenses;
 		next unless ($license);
 		next
-			if defined(
-			$coverage->get_range( $pos, $pos_license{$pos}{$license}->end )
-				->get_element(0) );
-		$coverage->set_range(
-			$pos_license{$pos}{$license}->begin,
-			$pos_license{$pos}{$license}->end,
-			$pos_license{$pos}{$license}
+			if $self->contains_tag( $pos, $pos_license{$pos}{$license}->[1] );
+		$self->tag(
+			Fulltext->new(
+				id    => [ $license, $license ],   # TODO: try resolve caption
+				begin => $pos_license{$pos}{$license}->[0],
+				end   => $pos_license{$pos}{$license}->[1]
+			)
 		);
 		$license{$license} = 1;
 	}
 
 	# grant, stepwise
+	my @prefixes;
 	$log->trace('scan stepwise for license grant');
 	foreach my $trait ( keys %{ $L{TRAITS_grant_prefix} } ) {
 
 		while ( $string =~ /$RE{"TRAIT_$trait"}/g ) {
-			next
-				if (
-				defined(
-					$coverage->get_range( $-[0], $+[0] )->get_element(0)
-				)
-				);
-			push @clues,
-				Trait->new(
-				name  => $trait,
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-				);
+			next if $self->contains_tag( $-[0], $+[0] );
+			push @prefixes, $self->note( $trait, $-[0], $+[0] );
 		}
 	}
 	LICENSED_UNDER:
-	foreach my $licensed_under (
-		sort { $a->end <=> $b->end }
-		grep { exists $L{TRAITS_grant_prefix}{ $_->name } } @clues
-		)
-	{
-		my $pos = $licensed_under->end;
+	foreach my $licensed_under ( sort { $a->[1] <=> $b->[1] } @prefixes ) {
+		my $pos = $licensed_under->[1];
 
 		# possible grant names
 		my @grant_types = (
@@ -693,34 +626,16 @@ method resolve
 		# scan for prepended version
 		substr( $string, $pos ) =~ $RE{ANCHORLEFT_NAMED_version};
 		if ( $+{version_number} ) {
-			push @clues,
-				Trait->new(
-				name  => 'version',
-				begin => $pos + $-[0],
-				end   => $pos + $+[0],
-				file  => $self,
-				);
+			$self->note( 'version', $pos + $-[0], $pos + $+[0] );
 			$version = $+{version_number};
 			if ( $+{version_later} ) {
-				push @clues,
-					Trait->new(
-					name  => 'or_later',
-					begin => $pos + $-[2],
-					end   => $pos + $+[2],
-					file  => $self,
-					);
+				$self->note( 'or_later', $pos + $-[2], $pos + $+[2] );
 				$later = $+{version_later};
 			}
 			if (
 				substr( $string, $pos + $+[0] ) =~ $L{LEFTANCHOR_version_of} )
 			{
-				push @clues,
-					Trait->new(
-					name  => 'version_of',
-					begin => $pos + $-[0],
-					end   => $pos + $+[0],
-					file  => $self,
-					);
+				$self->note( 'version_of', $pos + $-[0], $pos + $+[0] );
 				$pos += $+[0];
 				@grant_types = @L_type_versioned;
 			}
@@ -732,51 +647,33 @@ method resolve
 		# scan for name
 		foreach my $id (@RE_NAME) {
 			if ( substr( $string, $pos ) =~ $RE{"NAME_$id"} ) {
-				$match{$id}{name}{ $pos + $-[0] } = Trait->new(
-					name  => "name($id)",
-					begin => $pos + $-[0],
-					end   => $pos + $+[0],
-					file  => $self,
-				);
+				$match{$id}{name}{ $pos + $-[0] }
+					= $self->note( "name($id)", $pos + $-[0], $pos + $+[0] );
 			}
 		}
 
 		# pick longest matched license name
 		# TODO: include all of most specific type when more are longest
-		my @names = nsort_by { $match{$_}{name}{$pos}->end }
+		my @names = nsort_by { $match{$_}{name}{$pos}->[1] }
 		grep { $match{$_} and $match{$_}{name} and $match{$_}{name}{$pos} }
 			@grant_types;
 		my $name = pop @names;
 		if (    $name
 			and $match{$name}{name}{$pos}
-			and !defined(
-				$coverage->get_range( $pos, $match{$name}{name}{$pos}->end )
-					->get_element(0)
-			)
-			and ( !$skip_stepwise or $L_grant_atomic_incomplete{$name} )
-			)
+			and !$self->contains_tag( $pos, $match{$name}{name}{$pos}->[1] )
+			and ( !$skip_stepwise or $L_grant_atomic_incomplete{$name} ) )
 		{
-			my $pos_end = $pos = $match{$name}{name}{$pos}->end;
+			my $pos_end = $pos = $match{$name}{name}{$pos}->[1];
 
 			# may include version
 			if ( !$version and grep { $_ eq $name } @L_type_versioned ) {
 				substr( $string, $pos ) =~ $RE{ANCHORLEFT_NAMED_version};
 				if ( $+{version_number} ) {
-					push @clues, Trait->new(
-						name  => 'version',
-						begin => $pos + $-[0],
-						end   => $pos + $+[0],
-						file  => $self,
-					);
+					$self->note( 'version', $pos + $-[0], $pos + $+[0] );
 					$version = $+{version_number};
 					$pos_end = $pos + $+[1];
 					if ( $+{version_later} ) {
-						push @clues, Trait->new(
-							name  => 'or_later',
-							begin => $pos + $-[2],
-							end   => $pos + $+[2],
-							file  => $self,
-						);
+						$self->note( 'or_later', $pos + $-[2], $pos + $+[2] );
 						$later   = $+{version_later};
 						$pos_end = $pos + $+[2];
 					}
@@ -787,12 +684,7 @@ method resolve
 				substr( $string, $pos )
 					=~ $RE{ANCHORLEFT_NAMED_version_later};
 				if ( $+{version_later} ) {
-					push @clues, Trait->new(
-						name  => 'or_later',
-						begin => $pos + $-[1],
-						end   => $pos + $+[1],
-						file  => $self,
-					);
+					$self->note( 'or_later', $pos + $-[1], $pos + $+[1] );
 					$later   = $+{version_later};
 					$pos_end = $pos + $+[1];
 				}
@@ -804,22 +696,12 @@ method resolve
 			}
 			if ($later) {
 				my $latername = "${name}_or_later";
-				push @clues, Trait->new(
-					name  => $latername,
-					begin => $licensed_under->begin,
-					end   => $pos_end,
-					file  => $self,
-				);
-				$grant{$latername} = $clues[-1];
+				$self->note( $latername, $licensed_under->[0], $pos_end );
+				$grant{$latername} = 1;
 				next LICENSED_UNDER if grep { $grant{$_} } @RE_NAME;
 			}
-			$grant{$name} = Trait->new(
-				name  => "grant($name)",
-				begin => $licensed_under->begin,
-				end   => $pos_end,
-				file  => $self,
-			);
-			push @clues, $grant{$name};
+			$self->note( "grant($name)", $licensed_under->[0], $pos_end );
+			$grant{$name} = 1;
 		}
 	}
 
@@ -833,12 +715,8 @@ method resolve
 				qw(addr_fsf_franklin_steet addr_fsf_mass addr_fsf_temple))
 			{
 				if ( defined $+{$_} ) {
-					push @flaws, Flaw->new(
-						id    => $Regexp::Pattern::License::RE{$_},
-						begin => $-[0],
-						end   => $+[0],
-						file  => $self,
-					);
+					$self->tag(
+						Flaw->new( id => $_, begin => $-[0], end => $+[0] ) );
 				}
 			}
 		}
@@ -848,14 +726,8 @@ method resolve
 	# TODO: conditionally limit to AGPL/GPL/LGPL
 	foreach (@RE_EXCEPTION) {
 		if ( $string =~ $RE{"EXCEPTION_$_"} ) {
-			my $exception = Exception->new(
-				id    => $Regexp::Pattern::License::RE{$_},
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
-			$coverage->set_range( $-[0], $+[0], $exception );
-			push @exceptions, $exception;
+			$self->tag(
+				Exception->new( id => $_, begin => $-[0], end => $+[0] ) );
 		}
 	}
 
@@ -864,12 +736,8 @@ method resolve
 
 	# generated file
 	if ( $string =~ $RE{TRAIT_generated} ) {
-		push @flaws, Flaw->new(
-			id    => $Regexp::Pattern::License::RE{generated},
-			begin => $-[0],
-			end   => $+[0],
-			file  => $self,
-		);
+		$self->tag(
+			Flaw->new( id => 'generated', begin => $-[0], end => $+[0] ) );
 	}
 
 	# multi-licensing
@@ -880,16 +748,7 @@ method resolve
 	if ( grep { $match{$_}{name} } @lgpl ) {
 		$log->trace('scan for LGPL dual-license grant');
 		if ( $string =~ $L{multi_1} ) {
-			my $meta = Trait->new(
-				name  => 'grant(multi#1)',
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
-			$log->tracef(
-				'detected custom pattern multi#1: %s %s %s: %s',
-				'lgpl', $1, $2, $-[0]
-			);
+			$self->note( 'grant(multi#1)', $-[0], $+[0] );
 			push @multilicenses, 'lgpl', $1, $2;
 		}
 	}
@@ -899,10 +758,7 @@ method resolve
 	if ( grep { $match{$_}{name} } @gpl ) {
 		$log->trace('scan for GPL dual-license grant');
 		if ( $string =~ $L{multi_2} ) {
-			$log->tracef(
-				'detected custom pattern multi#2: %s %s %s: %s',
-				'gpl', $1, $2, $-[0]
-			);
+			$self->note( 'grant(multi#2)', $-[0], $+[0] );
 			push @multilicenses, 'gpl', $1, $2;
 		}
 	}
@@ -915,21 +771,12 @@ method resolve
 
 		# LGPL, dual versions last
 		if ( $string =~ $L{lgpl_5} ) {
-			my $grant = Trait->new(
-				name  => 'grant(lgpl#5)',
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
-			$license = "LGPL (v$1 or v$2) $license";
-			my $expr = "LGPL-$1 or LGPL-$2";
-			push @expressions,
+			$self->tag(
 				Grant->new(
-				name  => $expr,
-				begin => $grant->begin,
-				end   => $grant->end,
-				file  => $grant->file,
-				);
+					id    => [ "LGPL-$1 or LGPL-$2", "LGPL (v$1 or v$2)" ],
+					begin => $-[0], end => $+[0],
+				)
+			);
 			$match{ 'lgpl_' . $1 =~ tr/./_/r }{custom} = 1;
 			$match{ 'lgpl_' . $2 =~ tr/./_/r }{custom} = 1;
 			$match{lgpl}{custom} = 1;
@@ -940,95 +787,25 @@ method resolve
 	if ( grep { $match{$_}{name} } @gpl ) {
 		$log->trace('scan for GPL or LGPL dual-license grant');
 		if ( $string =~ $L{gpl_7} ) {
-			my $grant = Trait->new(
-				name  => "grant(gpl#7)",
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
+			$self->note( "grant(gpl#7)", $-[0], $+[0] );
 			$gen_license->( 'gpl', $1, $2, 'lgpl', $3, $4 );
 			$match{gpl}{custom}  = 1;
 			$match{lgpl}{custom} = 1;
 		}
 	}
 
-	# BSD
-	if ( grep { $match{$_}{name} } @L_contains_bsd
-		and $string =~ $L{bsd_1} )
-	{
-		$log->trace('scan for BSD fulltext');
-		my $grant = Trait->new(
-			name  => 'license(bsd#1)',
-			begin => $-[0],
-			end   => $+[0],
-			file  => $self,
-		);
-		for ($string) {
-			next if ( $license{bsd_4_clause} );
-			if ( $string =~ $RE{TRAIT_clause_advertising} ) {
-				my $grant = Trait->new(
-					name  => 'clause_advertising',
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
-				$gen_license->('bsd_4_clause');
-				next;
-			}
-			next if ( $license{bsd_3_clause} );
-			if ( $string =~ $RE{TRAIT_clause_non_endorsement} ) {
-				my $grant = Trait->new(
-					name  => 'clause_non_endorsement',
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
-				$gen_license->('bsd_3_clause');
-				next;
-			}
-			next if ( $license{bsd_2_clause} );
-			if ( $string =~ $RE{TRAIT_clause_reproduction} ) {
-				next
-					if (
-					defined(
-						$coverage->get_range( $-[0], $+[0] )->get_element(0)
-					)
-					);
-				my $grant = Trait->new(
-					name  => 'clause_reproduction',
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
-				$gen_license->('bsd_2_clause');
-				next;
-			}
-			$gen_license->('bsd');
-		}
-	}
-
 	# Apache dual-licensed with GPL/BSD/MIT
 	if ( $match{apache}{name} ) {
 		$log->trace('scan for Apache license grant');
 		for ($string) {
 			if ( $string =~ $L{apache_1} ) {
-				my $grant = Trait->new(
-					name  => 'grant(apache#1)',
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
+				$self->note( 'grant(apache#1)', $-[0], $+[0] );
 				$gen_license->( 'apache', $1, $2, 'gpl', $3, $4 );
 				$match{ $patterns2id->( 'apache', $1 ) }{custom} = 1;
 				next;
 			}
 			if ( $string =~ $L{apache_2} ) {
-				my $grant = Trait->new(
-					name  => 'grant(apache#2)',
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
+				$self->note( 'grant(apache#2)', $-[0], $+[0] );
 				$gen_license->(
 					'apache', $1, $2,
 					$3 ? "bsd_${3}_clause" : ''
@@ -1037,12 +814,7 @@ method resolve
 				next;
 			}
 			if ( $string =~ $L{apache_4} ) {
-				my $grant = Trait->new(
-					name  => 'grant(apache#4)',
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
+				$self->note( 'grant(apache#4)', $-[0], $+[0] );
 				$gen_license->( 'apache', $1, $2, 'mit', $3, $4 );
 				$match{ $patterns2id->( 'apache', $1 ) }{custom} = 1;
 				next;
@@ -1053,87 +825,55 @@ method resolve
 	# FSFUL
 	# FIXME: add test covering this pattern
 	$log->trace('scan for FSFUL fulltext');
-	if ( not $license{fsful} ) {
-		if ( $string =~ $L{fsful} ) {
-			my $grant = Trait->new(
-				name  => 'grant(fsful#1)',
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
-			$license = "FSF Unlimited ($1 derivation) $license";
-			my $expr = "FSFUL~$1";
-			push @expressions,
-				Fulltext->new(
-				name  => $expr,
-				begin => $grant->begin,
-				end   => $grant->end,
-				file  => $grant->file,
-				);
-			$match{fsful}{custom} = 1;
-		}
+	if (   !$license{fsful}
+		and $string =~ $L{fsful} )
+	{
+		$self->tag(
+			Fulltext->new(
+				id    => [ "FSFUL~$1", "FSF Unlimited ($1 derivation)" ],
+				begin => $-[0], end => $+[0],
+			)
+		);
+		$match{fsful}{custom} = 1;
 	}
 
 	# FSFULLR
 	# FIXME: add test covering this pattern
 	$log->trace('scan for FSFULLR fulltext');
-	if ( not $license{fsfullr} ) {
-		if ( $string =~ $L{fsfullr} ) {
-			my $grant = Trait->new(
-				name  => 'grant(fsfullr#1)',
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
-			$license
-				= "FSF Unlimited (with Retention, $1 derivation) $license";
-			my $expr = "FSFULLR~$1";
-			push @expressions,
-				Fulltext->new(
-				name  => $expr,
-				begin => $grant->begin,
-				end   => $grant->end,
-				file  => $grant->file,
-				);
-			$match{fsfullr}{custom} = 1;
-		}
+	if (   !$license{fsfullr}
+		and $string =~ $L{fsfullr} )
+	{
+		$self->tag(
+			Fulltext->new(
+				id => [
+					"FSFULLR~$1",
+					"FSF Unlimited (with Retention, $1 derivation)"
+				],
+				begin => $-[0], end => $+[0],
+			)
+		);
+		$match{fsfullr}{custom} = 1;
 	}
 
 	# usage
 	$log->trace('scan atomic for singleversion usage license grant');
 	foreach my $id (@L_type_usage) {
 		next if ( $match{$id}{custom} );
-		if ( !$grant{$id}
-			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic ) )
+		if (    !$grant{$id}
+			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic )
+			and $string =~ $RE{"GRANT_$id"} )
 		{
-			if ( $string =~ $RE{"GRANT_$id"} ) {
-				my $grant = Trait->new(
-					name  => "grant($id)",
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
-				unless (
-					defined(
-						$coverage->get_range( $-[0], $+[0] )->get_element(0)
-					)
-					)
-				{
-					$grant{$id} = Grant->new(
-						name  => $id,
-						begin => $grant->begin,
-						end   => $grant->end,
-						file  => $grant->file,
-					);
-				}
+			if ( $self->contains_tag( $-[0], $+[0] ) ) {
+				$log->tracef( 'skip grant in covered range: %s', $id );
+			}
+			else {
+				$self->tag(
+					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
+				$grant{$id} = 1;
 			}
 		}
 
 		if ( $grant{$id} ) {
-			$coverage->set_range(
-				$grant{$id}->begin, $grant{$id}->end,
-				$grant{$id}
-			);
 			$gen_license->( $id2patterns->($id) );
 
 			# skip singleversion and unversioned equivalents
@@ -1157,36 +897,20 @@ method resolve
 		if (    !$license{$id}
 			and !$grant{$id}
 			and !$match{$id}{custom}
-			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic ) )
+			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic )
+			and $string =~ $RE{"GRANT_$id"} )
 		{
-			if ( $string =~ $RE{"GRANT_$id"} ) {
-				my $grant = Trait->new(
-					name  => "grant($id)",
-					begin => $-[0],
-					end   => $+[0],
-					file  => $self,
-				);
-				unless (
-					defined(
-						$coverage->get_range( $-[0], $+[0] )->get_element(0)
-					)
-					)
-				{
-					$grant{$id} = Grant->new(
-						name  => $id,
-						begin => $grant->begin,
-						end   => $grant->end,
-						file  => $grant->file,
-					);
-				}
+			if ( $self->contains_tag( $-[0], $+[0] ) ) {
+				$log->tracef( 'skip grant in covered range: %s', $id );
+			}
+			else {
+				$self->tag(
+					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
+				$grant{$id} = 1;
 			}
 		}
 
 		if ( $license{$id} or $grant{$id} ) {
-			$coverage->set_range(
-				$grant{$id}->begin, $grant{$id}->end,
-				$grant{$id}
-			) if $grant{$id};
 			$gen_license->( $id2patterns->($id) )
 				unless ( $match{$id}{custom} );
 
@@ -1203,50 +927,26 @@ method resolve
 	foreach my $id (@L_type_versioned) {
 		next if ( $match{$id}{custom} );
 
-		# skip name part of another name detected as grant
-		# TODO: use less brittle method than name of clue
-		next
-			if ( $id eq 'cc_by'
-			and grep { $_->name eq 'grant(cc_by_sa_3)' } @clues );
-
 		# skip embedded or referenced licenses
 		next if ( $license{rpsl_1} and grep { $id eq $_ } qw(mpl python) );
 
 		next if ( $license{$id} );
-		if ( !$grant{$id}
-			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic ) )
+		if (    !$grant{$id}
+			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic )
+			and $RE{"GRANT_$id"}
+			and $string =~ $RE{"GRANT_$id"} )
 		{
-			if ( $RE{"GRANT_$id"} ) {
-				if ( $string =~ $RE{"GRANT_$id"} ) {
-					my $grant = Trait->new(
-						name  => "grant($id)",
-						begin => $-[0],
-						end   => $+[0],
-						file  => $self,
-					);
-					unless (
-						defined(
-							$coverage->get_range( $-[0], $+[0] )
-								->get_element(0)
-						)
-						)
-					{
-						$grant{$id} = Grant->new(
-							name  => $id,
-							begin => $grant->begin,
-							end   => $grant->end,
-							file  => $grant->file,
-						);
-					}
-				}
+			if ( $self->contains_tag( $-[0], $+[0] ) ) {
+				$log->tracef( 'skip grant in covered range: %s', $id );
+			}
+			else {
+				$self->tag(
+					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
+				$grant{$id} = 1;
 			}
 		}
 
 		if ( $grant{$id} ) {
-			$coverage->set_range(
-				$grant{$id}->begin, $grant{$id}->end,
-				$grant{$id}
-			);
 			$gen_license->($id);
 		}
 	}
@@ -1274,51 +974,43 @@ method resolve
 			and !$grant{$id}
 			and $string =~ $RE{"GRANT_$id"} )
 		{
-			my $grant = Trait->new(
-				name  => "grant($id)",
-				begin => $-[0],
-				end   => $+[0],
-				file  => $self,
-			);
-			unless (
-				defined(
-					$coverage->get_range( $-[0], $+[0] )->get_element(0)
-				)
-				)
-			{
-				$grant{$id} = Grant->new(
-					name  => $id,
-					begin => $grant->begin,
-					end   => $grant->end,
-					file  => $grant->file,
-				);
+			if ( $self->contains_tag( $-[0], $+[0] ) ) {
+				$log->tracef( 'skip grant in covered range: %s', $id );
+			}
+			else {
+				$self->tag(
+					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
+				$grant{$id} = 1;
 			}
 		}
 		if ( $license{$id} or $grant{$id} ) {
-			$coverage->set_range(
-				$grant{$id}->begin, $grant{$id}->end,
-				$grant{$id}
-			) if $grant{$id};
 			$gen_license->($id);
 		}
 	}
 
-	$license =~ s/$L{trailing_space}//;
-	$expr = join( ' and/or ', sort map { $_->name } @expressions );
+	# Expressions and exceptions contain DEP-5 or SPDX identifiers;
+	# flaws contains non-SPDX notes.
+	my ( $licenses, $exceptions, $flaws ) = $self->get_tags;
+
+	my @expressions = map { $_->[1] } @$licenses;
+	my @license     = map { $_->[2] } @$licenses;
+	$expr    = join( ' and/or ', sort @expressions );
+	$license = join( ' and/or ', sort @license );
 	$expr    ||= 'UNKNOWN';
 	$license ||= 'UNKNOWN';
-	if (@exceptions) {
+
+	if (@$exceptions) {
 		$expr = "($expr)"
 			if ( @expressions > 1 );
 		$expr .= ' with ' . join(
 			'_AND_',
-			sort map { $self->best_value( $_->id, 'name' ) } @exceptions
+			sort map { $_->[1] } @$exceptions
 		) . ' exception';
 	}
-	if (@flaws) {
+	if (@$flaws) {
 		$license .= ' [' . join(
 			', ',
-			sort map { $self->best_value( $_->id, qw(caption name) ) } @flaws
+			sort map { $_->[2] } @$flaws
 		) . ']';
 	}
 	$log->infof( 'resolved license expression: %s', $expr );
@@ -1334,7 +1026,7 @@ or with scheme-less naming as a short description.
 
 =cut
 
-method as_text
+method as_text ()
 {
 	if ( $naming->list_schemes ) {
 		$self->resolve
