@@ -1,8 +1,10 @@
-use Feature::Compat::Class 0.04;
-
-use v5.12;
+use v5.20;
 use utf8;
 use warnings;
+use feature qw(signatures);
+no warnings qw(experimental::signatures);
+
+use Feature::Compat::Class 0.04;
 
 =head1 NAME
 
@@ -10,7 +12,7 @@ String::License::Naming::SPDX - licenses as named by SPDX
 
 =head1 VERSION
 
-Version v0.0.5
+Version v0.0.6
 
 =head1 SYNOPSIS
 
@@ -43,7 +45,7 @@ or as fallback by an internal name.
 
 =cut
 
-package String::License::Naming::SPDX v0.0.5;
+package String::License::Naming::SPDX v0.0.6;
 
 use Carp            qw(croak);
 use Log::Any        ();
@@ -93,7 +95,7 @@ Returns a list of license naming schemes in use.
 
 =cut
 
-method list_schemes
+method list_schemes ()
 {
 	return @$schemes;
 }
@@ -105,7 +107,7 @@ each labeled by SPDX shortname.
 
 =cut
 
-method list_licenses
+method list_licenses ()
 {
 	return String::License::Naming::resolve_shortnames( $keys, $schemes );
 }
